 /***************************************************************************\
|*                                                                           *|
|*        Copyright (c) 1999 NVIDIA, Corp.  All rights reserved.             *|
|*                                                                           *|
|*     NOTICE TO USER:   The source code  is copyrighted under  U.S. and     *|
|*     international laws.   NVIDIA, Corp. of Sunnyvale, California owns     *|
|*     the copyright  and as design patents  pending  on the design  and     *|
|*     interface  of the NV chips.   Users and possessors of this source     *|
|*     code are hereby granted  a nonexclusive,  royalty-free  copyright     *|
|*     and  design  patent license  to use this code  in individual  and     *|
|*     commercial software.                                                  *|
|*                                                                           *|
|*     Any use of this source code must include,  in the user documenta-     *|
|*     tion and  internal comments to the code,  notices to the end user     *|
|*     as follows:                                                           *|
|*                                                                           *|
|*     Copyright (c) 1993-1997  NVIDIA, Corp.    NVIDIA  design  patents     *|
|*     pending in the U.S. and foreign countries.                            *|
|*                                                                           *|
|*     NVIDIA, CORP.  MAKES  NO REPRESENTATION ABOUT  THE SUITABILITY OF     *|
|*     THIS SOURCE CODE FOR ANY PURPOSE.  IT IS PROVIDED "AS IS" WITHOUT     *|
|*     EXPRESS OR IMPLIED WARRANTY OF ANY KIND.  NVIDIA, CORP. DISCLAIMS     *|
|*     ALL WARRANTIES  WITH REGARD  TO THIS SOURCE CODE,  INCLUDING  ALL     *|
|*     IMPLIED   WARRANTIES  OF  MERCHANTABILITY  AND   FITNESS   FOR  A     *|
|*     PARTICULAR  PURPOSE.   IN NO EVENT SHALL NVIDIA, CORP.  BE LIABLE     *|
|*     FOR ANY SPECIAL, INDIRECT, INCIDENTAL,  OR CONSEQUENTIAL DAMAGES,     *|
|*     OR ANY DAMAGES  WHATSOEVER  RESULTING  FROM LOSS OF USE,  DATA OR     *|
|*     PROFITS,  WHETHER IN AN ACTION  OF CONTRACT,  NEGLIGENCE OR OTHER     *|
|*     TORTIOUS ACTION, ARISING OUT  OF OR IN CONNECTION WITH THE USE OR     *|
|*     PERFORMANCE OF THIS SOURCE CODE.                                      *|
|*                                                                           *|
 \***************************************************************************/

/*
 * Os interface definitions needed by os-interface.c
 */

#ifndef _OS_INTERFACE_H_
#define _OS_INTERFACE_H_

/******************* Operating System Interface Routines *******************\
*                                                                           *
* Module: os-interface.h                                                    *
*       Included by os.h                                                    *
*       Operating system wrapper functions used to abstract the OS.         *
*                                                                           *
\***************************************************************************/

/*
 * Define away Microsoft compiler extensions when possible
 */

#define __stdcall
#define far
#define PASCAL

/*
 * misc types used within resman
 */

typedef void *        HANDLE;
typedef void *        FLATPTR;
typedef unsigned long DWORD;

typedef          char      CHAR;
typedef          int       INT;
typedef          int       SHORT;
typedef          char     *PCHAR;
typedef          int      *PSHORT;
typedef          void     *PVOID;
typedef          char far *LPCHAR;
typedef          char far *LPSTR;
typedef          int  far *LPSHORT;
typedef unsigned char      UCHAR;
typedef unsigned int       USHORT;
typedef unsigned long      ULONG;
typedef unsigned char     *PUCHAR;
typedef unsigned int      *PUSHORT;
typedef unsigned long     *PULONG;
typedef unsigned char far *LPUCHAR;
typedef unsigned int  far *LPUSHORT;
typedef unsigned long far *LPULONG;

//---------------------------------------------------------------------------
//
//  Function prototypes for OS interface.
//
//---------------------------------------------------------------------------

RM_STATUS   osAllocMem              (void **, U032);
RM_STATUS   osFreeMem               (void *);
RM_STATUS   osGetCurrentTime        (U032 * seconds, U032 *useconds);
RM_STATUS   osDelay                 (U032);
U032        osGetCpuFrequency       (void);
RM_STATUS   osKillProcess(U032 pid, U032 signal);
char*       osStringCopy            (char *, const char *);
RM_STATUS   osStringNCopyFromUser(char *, const char *, int n);
U032        osStringLength          (const char *);
void       *osMemCopy(void *, const void *, U032);
RM_STATUS   osMemCopyFromUser(void *, const void *, U032);
RM_STATUS   osMemCopyToUser(void *, const void *, U032);
void       *osMemSet(void *, int c, U032);
int         osMemCmp(const void *buf0, const void *buf1, U032);
S032        osStringCompare         (const char *, const char *);

U032        osPciInitHandle(U008 bus, U008 slot, U008 function, U016 *vendor, U016 *device);
U008        osPciReadByte(U032 handle, U008 offset);
U016        osPciReadWord(U032 handle, U008 offset);
U032        osPciReadDword(U032 handle, U008 offset);
VOID        osPciWriteByte(U032 handle, U008 offset, U008 value);
VOID        osPciWriteWord(U032 handle, U008 offset, U016 value);
VOID        osPciWriteDword(U032 handle, U008 offset, U032 value);

/* remap IO memory to kernel space */
void *osMapKernelSpace(unsigned long start, unsigned long size, U008 cached);
void osUnmapKernelSpace(void *addr, unsigned long size);

/* remap memory to user space */
void *osMapUserSpace(void *kaddr, void **priv, unsigned int size_bytes, U008 cached);
void osUnmapUserSpace(void *uaddr, void *priv);

// like osAllocMem, but no debugging info written into pages
// needed for NT, but only pass-through under Linux
RM_STATUS osAllocPool(void **addr, unsigned int size_bytes);
RM_STATUS osFreePool(void *addr);


void *osAllocContigPages(unsigned long num_pages);
void osFreeContigPages(void *addr); 
U032 osGetKernPhysAddr(void *kern_virt_addr);


RM_STATUS   osReadRegistryDword     (PHWINFO, char*, char*, U032*);
RM_STATUS   osWriteRegistryDword    (PHWINFO, char*, char*, U032);
RM_STATUS   osReadRegistryBinary    (PHWINFO, char*, char*, U008*, U032 *);
RM_STATUS   osWriteRegistryBinary   (PHWINFO, char*, char*, U008*, U032);

BOOL osPciDevicePresent(unsigned short vendor, unsigned short device);

void osIoWriteByte(unsigned long address, unsigned char value);
void osIoWriteWord(unsigned long address, unsigned short value);
void osIoWriteDword(unsigned long address, unsigned int value);

unsigned char  osIoReadByte(unsigned long address);
unsigned short osIoReadWord(unsigned long address);
unsigned int   osIoReadDword(unsigned long address);

unsigned int osCli(unsigned int flags);
unsigned int osSti(unsigned int flags);

#define osDisableHardwareInterrupts(pDev)   flags = osCli(flags)
#define osEnableHardwareInterrupts(pDev)    flags = osSti(flags)

RM_STATUS osCallVideoBIOS(PHWINFO, U032*, U032*, U032*, U032*, VOID*);

/*
 * misc routines; defined in misc/asm.c
 */

extern unsigned long long nv_rdtsc(void);
extern unsigned int nv_rdcr4(void);
extern int nv_cpuid(int op, U032 *eax, U032 *ebx, U032 *ecx, U032 *edx);
extern int nv_rdmsr(int addr, U032 *eax, U032 *edx);


/*
 * nv_error(), nv_message(), nv_panic() support
 */

#define NV_ERROR_ERRNO     (1<<31)     /* hi bit; use 'errno' */
#define NV_ERROR_PANIC     (1<<30)     /* err is fatal; panic */
#define NV_ERROR_MESSAGE (1<<29)     /* just msg; not error */
#define NV_ERROR_STATUS   (1<<28)     /* rtems status in error */
#define NV_ERROR_WARNING (1<<27)     /* just warning; no error */

#define NV_ERROR_MASK (  \
                            NV_ERROR_ERRNO | NV_ERROR_PANIC | \
                            NV_ERROR_MESSAGE | \
                            NV_ERROR_WARNING | \
                            NV_ERROR_STATUS) /* all */


int  nv_error(int error_code, const char *printf_format, ...);
void nv_panic(const char *printf_format, ...) __attribute__ ((noreturn));
void nv_abort(const char *printf_format, ...) __attribute__ ((noreturn));
void nv_message(const char *printf_format, ...);
void nv_warning(const char *printf_format, ...);
void nv_error_new_failed(void);
int nv_printf(const char *printf_format, ...);
int nv_debug(int level, const char *printf_fmt, ...);


//---------------------------------------------------------------------------
//
//  Debug macros.
//
//---------------------------------------------------------------------------

// enable debugging if any OS debugging flag is set
#undef DEBUGGING
#if defined(DEBUG) || defined(DBG)
#define DEBUGGING
#endif


// Debug level values
#define DEBUGLEVEL_TRACEINFO	0     // For informational debug trace info
#define DEBUGLEVEL_SETUPINFO	1     // For informational debug setup info
#define DEBUGLEVEL_USERERRORS	2     // For debug info on app level errors
#define DEBUGLEVEL_WARNINGS	    3     // For RM debug warning info
#define DEBUGLEVEL_ERRORS	    4     // For RM debug error info

void    osDbgInit(void);
void 	osDbgBreakPoint(void);

/*
 * os Debug prints.  Not sure why we have these in addition to DBG_PRINT__
 * macros. Define them all in terms of 'nv_debug()' which is a general printf
 * Currently on the UNIX port, these will compile even on non-debug
 */

#define osDbgPrintString(level,str)   nv_debug(level, "%s", str)
#define osDbgPrintCR(level)               nv_debug(level, "\n")
#define osDbgPrintValue(level,n)       nv_debug(level, "0x%x", n)
#define osDbgPrintStringValue(level,str,n) nv_debug(level, "%s 0x%x\n", str, n)
#define osDbgPrintStringDec(level,str,n)  nv_debug(level, "%s %d\n", str, n)
#define osDbgPrintStringPtr(level,str,ptr) nv_debug(level, "%s 0x%x\n", str, ptr)
#define osDbgPrintStringVal64(level,str,ptr) nv_debug(level, "%s FIXME!! no 64 bit support yet!!\n", str, ptr)

#ifdef DEBUGGING
#define DBG_BREAKPOINT()      			osDbgBreakPoint()
#define DBG_PRINT_STRING(d,s)  			osDbgPrintString((d),(s))
#define DBG_PRINT_VALUE(d,v)  			osDbgPrintValue((d),(v))
#define DBG_PRINT_CR(d,v)  				osDbgPrintCR((d))
#define DBG_PRINT_STRING_VALUE(d,s,v)	osDbgPrintStringValue((d),(s),(v))
#define DBG_PRINT_STRING_DEC(d,s,v)		osDbgPrintStringDec((d),(s),(v))
#define DBG_PRINT_STRING_PTR(d,s,v)		osDbgPrintStringPtr((d),(s),(v))
#define DBG_PRINT_STRING_VAL64(d,s,v)           osDbgPrintStringVal64((d),(s),(v))
#define DBG_PRINTF(level, fmt, args...)                 nv_debug(level, fmt, ## args)

#define DBG_VAL_PTR(p)  \
do { \
    U032 *ptr = (void *) (p); \
    if (ptr[-1] != NV_MARKER1){ \
        DBG_PRINTF(DEBUGLEVEL_ERRORS, "NVRM: Invalid pointer!\n"); \
        DBG_PRINTF(DEBUGLEVEL_ERRORS, "%s: line %d\n", __FILE__, __LINE__); \
        DBG_BREAKPOINT(); \
    } \
    /* pick out size and check last word */    \
    /* note using '/' to clear out V or K malloc type bits */ \
    else if (ptr[ptr[-2]/4 - 3] != NV_MARKER2) { \
        DBG_PRINTF(DEBUGLEVEL_ERRORS, "NVRM: Trashed buffer end!\n"); \
        DBG_PRINTF(DEBUGLEVEL_ERRORS, "%s: line %d\n", __FILE__, __LINE__); \
        DBG_BREAKPOINT(); \
    } \
} while(0)
#else 

// no debugging -- define out debug macros
#define DBG_BREAKPOINT()
#define DBG_PRINT_STRING(d,s)
#define DBG_PRINT_VALUE(d,v)
#define DBG_PRINT_CR(d,v)
#define DBG_PRINT_STRING_VALUE(d,s,v)
#define DBG_PRINT_STRING_DEC(d,s,v)
#define DBG_PRINT_STRING_PTR(d,s,v)
#define DBG_PRINT_STRING_VAL64(d,s,v)
#define DBG_VAL_PTR(p)

#define DBG_PRINTF(level, fmt, args...)

#endif // DEBUGGING

/* AGPGART */
BOOL KernInitAGP(VOID **ap_phys_base, VOID **ap_mapped_base, U032 *apsize);
BOOL KernTeardownAGP();
RM_STATUS KernAllocAGPPages(VOID **pAddress, U032 PageCount, VOID **priv_data);
RM_STATUS KernFreeAGPPages(VOID **pAddress, VOID  *priv_data);



#endif // _OS_INTERFACE_H_
